# -*- coding: utf-8 -*-
from os import listdir
from os.path import isfile, join
import csv
import pandas as pd
import re
import pickle
from sklearn.feature_extraction.text import TfidfVectorizer, CountVectorizer, TfidfTransformer
from sklearn.model_selection import StratifiedKFold
from sklearn.metrics import accuracy_score
import os


import numpy as np
from dl_architecture import make_charvec, build_model, build_general_model
from sklearn.preprocessing import Normalizer
from sklearn import pipeline
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.pipeline import FeatureUnion
from sklearn import preprocessing
from sklearn.metrics import f1_score

import gc
import string
import math

from collections import defaultdict
from keras import backend as K
from keras.utils.conv_utils import convert_kernel



def remove_email(text, replace_token):
    return re.sub(r'[\w\.-]+@[\w\.-]+', replace_token, text)


def remove_url(text, replace_token):
    regex = 'http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+'
    return re.sub(regex, replace_token, text)


def preprocess_data(df_data, column, tags_to_idx):

    #print("Data shape before preprocessing:", df_data.shape)
    df_data['text_clean'] = df_data[column].map(lambda x: remove_url(x, "HTTPURL"))
    df_data['text_clean'] = df_data['text_clean'].map(lambda x: remove_email(x, 'EMAIL'))

    #print("Tags to idx: ", tags_to_idx)
    #print("Preprocessed data columns: ", df_data.columns)

    return df_data


class text_col(BaseEstimator, TransformerMixin):
    def __init__(self, key):
        self.key = key
    def fit(self, x, y=None):
        return self
    def transform(self, data_dict):
        return data_dict[self.key]



def predict(data_test, column, lang, weights_path, data_path):

    textmodel_data = pickle.load(open(data_path, 'rb'))

    unigrams_shape, num_classes, charvec_shape, char_vocab_size, feature_union, char_vocab, max_train_len_char, tags_to_idx = textmodel_data
    #print(tags_to_idx)
    xtest = preprocess_data(data_test, column, tags_to_idx=tags_to_idx)
    feature_union_test = feature_union.transform(xtest)
    #print(unigrams_shape, num_classes, charvec_shape, char_vocab_size, feature_union_test.shape, len(char_vocab), max_train_len_char, tags_to_idx)

    charvec_test, _, _ = make_charvec(xtest.text_clean.tolist(), train=False, char_vocab=char_vocab, max_text_len=max_train_len_char)
    #print(charvec_test.shape)


   

    if lang != 'all':
        model = build_model(unigrams_shape, num_classes, charvec_shape, char_vocab_size)
    else:
        model = build_general_model(unigrams_shape, num_classes, charvec_shape, char_vocab_size)
    model.load_weights(weights_path)
    predictions = model.predict([feature_union_test, charvec_test]).argmax(axis=-1)
    return predictions, tags_to_idx
    


import numpy as np
from keras.layers import Conv1D
from keras.models import Model
from keras.layers import Input, Embedding, Dense, Dropout, Concatenate, BatchNormalization, MaxPooling1D, Flatten
from keras.layers.merge import concatenate



def build_model(unigrams_shape, num_classes, charvec_shape, char_vocab_size):
    
    # build model
    inputs = []

    #feature union
    feature_union = Input(shape=(unigrams_shape,))
    inputs.append(feature_union)

    # char level
    char_sequence_input = Input(shape=(charvec_shape,), dtype='int32')
    char_dense = charConv(char_sequence_input, charvec_shape, char_vocab_size)
    inputs.append(char_sequence_input)

    mergedconv = concatenate([feature_union, char_dense])
    dense_all = Dense(256, activation='relu')(mergedconv)
    dropout = Dropout(0.4)(dense_all)
    dense3 = Dense(num_classes, activation='softmax')(dropout)
    model = Model(inputs=inputs, outputs=dense3)
    model.summary()
    model.compile(loss='sparse_categorical_crossentropy',
                  optimizer="adam",
                  metrics=['acc'])
    return model


def build_general_model(unigrams_shape, num_classes, charvec_shape, char_vocab_size):
    
    # build model
    inputs = []

    # feature union
    feature_union = Input(shape=(unigrams_shape,))
    inputs.append(feature_union)

    # char level
    char_sequence_input = Input(shape=(charvec_shape,), dtype='int32')
    char_dense = charConvGeneral(char_sequence_input, charvec_shape, char_vocab_size)
    inputs.append(char_sequence_input)
    
    mergedconv = concatenate([feature_union, char_dense])
    dense_all = Dense(256, activation='relu')(mergedconv)
    dropout = Dropout(0.4)(dense_all)
    dense3 = Dense(num_classes, activation='softmax')(dropout)
    model = Model(inputs=inputs, outputs=dense3)
    model.summary()
    model.compile(loss='sparse_categorical_crossentropy',
                  optimizer="adam",
                  metrics=['acc'])
    return model


def make_charvec(data, train=True, char_vocab={}, max_text_len=0):
    counter = 2
    if train:
        all_texts = ' '.join(data)
        for char in all_texts:
            if char not in char_vocab:
                char_vocab[char] = counter
                counter += 1
        max_text_len = max([len(text) for text in data])
    else:
        if char_vocab == [] or max_text_len == 0:
            raise Exception('You have to add char vocabulary and max train text len as input')
    output_chars = np.zeros((len(data), max_text_len), dtype='int32')
    for i in range(len(data)):
        for j in range(max_text_len):
            if j < len(data[i]):
                if data[i][j] in char_vocab:
                    output_chars[i][j] = char_vocab[data[i][j]]
                else:
                    output_chars[i][j] = 1

    return output_chars, char_vocab, max_text_len


def charConv(sequence_input, max_seq_length, vocab_size):
    convs = []
    filter_sizes = [4, 5]


    embedding_layer = Embedding(vocab_size, 200, input_length=max_seq_length,)(sequence_input)

    for fsz in filter_sizes:
        l_conv = Conv1D(filters=172, kernel_size=fsz, activation='relu')(embedding_layer)
        l_bn = BatchNormalization()(l_conv)
        l_pool = MaxPooling1D(fsz)(l_bn)
        convs.append(l_pool)

    l_merge = Concatenate(axis=1)(convs)
    l_cov1 = Conv1D(200, 5, activation='relu')(l_merge)
    l_bn1 = BatchNormalization()(l_cov1)
    l_pool1 = MaxPooling1D(40)(l_bn1)
    l_dropout = Dropout(0.4)(l_pool1)
    l_flat = Flatten()(l_dropout)
    return l_flat


def charConvGeneral(sequence_input, max_seq_length, vocab_size):
    convs = []
    filter_sizes = [4, 5]
    embedding_layer = Embedding(vocab_size, 200, input_length=max_seq_length, )(sequence_input)

    for fsz in filter_sizes:
        l_conv = Conv1D(filters=172, kernel_size=fsz, activation='relu')(embedding_layer)
        l_bn = BatchNormalization()(l_conv)
        l_pool = MaxPooling1D(fsz)(l_bn)
        convs.append(l_pool)

    l_merge = Concatenate(axis=1)(convs)
    l_cov1 = Conv1D(200, 5, activation='relu')(l_merge)
    l_bn1 = BatchNormalization()(l_cov1)
    l_pool1 = MaxPooling1D(40)(l_bn1)
    l_dropout = Dropout(0.4)(l_pool1)
    l_flat = Flatten()(l_dropout)
    return l_flat
